{-# LANGUAGE CPP #-}
{-
parmap and mapReduce skeletons
-}

module ParMap where

import System.IO.Unsafe
import Control.Monad
import Control.Concurrent
import Data.List

import Control.Parallel.Eden

------------------------------------
-- Eden's parallel map
parMapEden :: (Trans a, Trans b) => (a -> b) -> [a] -> [b]
parMapEden = spawn . repeat . process 

--------------------------------------
-- Eden process farm (np processes, each working on a whole sublist)
farm0 :: (Trans a, Trans b) => (a -> b) -> [a] -> [b]
farm0 f xs = join results
    where results = spawn (repeat (process (map f))) parts
          parts   = distribute noPe xs -- noPe, so use all nodes
          ----------------------------
          -- join . distribute n == id
          join    =  error "you write one"
          distribute n = error "you write it"

---------------------------------------
parMapWP :: (Trans a, Trans b) => (a -> b) -> [a] -> [b]
parMapWP = edenWP (max 1 (noPe-1)) 30 -- hard-coded prefetch here

edenWP :: (Trans t, Trans r) =>
            Int ->			-- #workers
            Int ->			-- prefetch
            (t -> r) ->                 -- worker
            [t] -> [r]
edenWP np prefetch f tasks = results
  where fromWorkers        = spawn workerProcs toWorkers
        workerProcs        = [process (zip [n,n..] . map f) | n<-[1..np]]
        toWorkers          = distribute tasks requests
        (newReqs, results) = (unzip . merge) fromWorkers
        requests           = initialReqs ++ newReqs
        initialReqs        = concat (replicate prefetch [1..np])
        distribute :: [t] -> [Int] -> [[t]]
        distribute tasks reqs = [taskList reqs tasks n | n<-[1..np]] 
          where taskList (r:rs) (t:ts) pe | pe == r    = t:(taskList rs ts pe)
                                          | otherwise  =    taskList rs ts pe
                taskList _      _      _  = []

---------------------------------------

parmapReduceStream :: (Trans a, Trans b) =>
                      Int -> -- no. of processes
                     (a -> b) -> -- mapped on input
                     (b -> b -> b) -> -- reduction (assumed commutative)
                     b -> -- neutral element for reduction
                     [a] -> b
parmapReduceStream np mapF redF neutral list
    = let sublists = unshuffle np list
          subFold = foldl' redF neutral . (map mapF)
          subRs = spawnF (repeat subFold) sublists
      in foldl redF neutral subRs

unshuffle :: Int -> [a] -> [[a]]
unshuffle 0 xs = [xs]
unshuffle 1 xs = [xs]
unshuffle n xs = [takeEach n (drop i xs) | i <- [0..n-1]]
 where takeEach n [] = []
       takeEach n (x:xs) = x : takeEach n (drop (n-1) xs)
